@@ -1,6 +1,9 @@
 #ifndef __GLOBALS_H__
 #define __GLOBALS_H__
 
+// XX should be named kextd_globals.h or similar
+// currently not suitable for sharing with other tools
+
 #include <CoreFoundation/CoreFoundation.h>
 #include <IOKit/IOKitLib.h>
 #include <libc.h>
@@ -10,6 +13,8 @@
 #include "PTLock.h"
 
 // in main.c
+void kextd_handle_sighup(int);
+
 extern char * progname;
 extern int g_verbose_level;
 
@@ -18,13 +23,19 @@ extern char * g_patch_dir;
 extern char * g_symbol_dir;
 extern Boolean gOverwrite_symbols;
 
+extern Boolean g_safe_boot_mode;
+
+extern Boolean gStaleStartupMkext;
+extern Boolean gStaleKernel;
+
 extern mach_port_t g_io_master_port;
 
 extern KXKextManagerRef gKextManager;
+
 extern CFRunLoopRef gMainRunLoop;
 extern CFRunLoopSourceRef gKernelRequestRunLoopSource;
 extern CFRunLoopSourceRef gRescanRunLoopSource;
-extern CFRunLoopSourceRef gClientRequestRunLoopSource;
+//extern CFRunLoopSourceRef gClientRequestRunLoopSource;       // XX not global?
 extern CFRunLoopSourceRef gCurrentNotificationRunLoopSource;
 
 extern PTLockRef gKernelRequestQueueLock;
